/*
 *  (c) 2006 by VWP / Oxy
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------
 */


package com.vwp.sound.mod.modplay.loader;

import com.vwp.sound.mod.modplay.module.ModuleUnits;

/**
 * @author torkjel
 * STM actually has a sane way of specifying the sample rate / note 
 * relationship. The sample decides what rate the sample should be played at 
 * when played as a c-3 note. The rest follows from that, by fundamental note 
 * theory. (The rate of one note is 2^(1/12) times higher than the previous one)
 */
public class STMUnits implements ModuleUnits {

    private static final String NAME = "stm";
    
    private double c3hz; 
    private static final double root2_12 = 1.059463094; 
    private static final double c3 = 12*4;
    
    public STMUnits(int c3hz) {
        this.c3hz = c3hz;
    }
    
    public double period2note(double period) {
        return rate2note(period);
    }

    public double note2period(double period) {
        return note2rate(period);
    }
    
    public double note2rate(double note) {
        double diff = note - c3;
        return c3hz * Math.pow(root2_12,diff); 
    }

    public double rate2note(double rate) {
        return Math.log(rate/c3hz)/Math.log(root2_12) + c3;
    }

    public double addPeriod(double note, double period) {
        return period2note(note2period(note) + period); 
    }

    public double getUpperNoteLimit() {
        return 10*12;
    }
    
    public double getLowerNoteLimit() {
        return 0;
    }

    public String getName() {
        return NAME;
    }
}
