/*
 *  (c) 2006 by VWP / Oxy
 *  (c) by Torkjel Hongve
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------
 */

package com.vwp.sound.mod.modplay.player.autoeffect;

import com.vwp.sound.mod.modplay.player.TrackState;

/**
 * Tiis class is necessery because the envelopes of an .XM module doesn't operate 
 * entirely independently. The only real work done by this class is to, in the 
 * constructor, test if the volume envelope is active and, if so, activate
 * the fadeout effect. All other calls are passed on to the individual effects. 
 * This should really be reimplemented as a test in Fadeout instead....
 * @author torkjel
 */
public class XmAutoEffects implements Envelope
{
	private VolumeEnvelope volEnv;
	private PanningEnvelope panEnv;
	private AutoVibrato autoVib;
	private Fadeout fadeout;

	public XmAutoEffects(
		VolumeEnvelope volEnv,
		PanningEnvelope panEnv,
		AutoVibrato autoVib,
		Fadeout fadeout)
	{
		this.volEnv = volEnv;
		this.panEnv = panEnv;
		this.autoVib = autoVib;
		this.fadeout = fadeout;
		if (volEnv.isActive())
			fadeout.activate();
		else
			fadeout.deActivate();
	}

	public void setNumberOfTracks(int tracks)
	{
		volEnv.setNumberOfTracks(tracks);
		panEnv.setNumberOfTracks(tracks);
		fadeout.setNumberOfTracks(tracks);
        autoVib.setNumberOfTracks(tracks);
	}

	public void doEffect(TrackState state, int track)
	{
		volEnv.doEffect(state, track);
		panEnv.doEffect(state, track);
		fadeout.doEffect(state, track);
        autoVib.doEffect(state, track);
	}

    public void keyOff(int track) {
        volEnv.keyOff(track);
        panEnv.keyOff(track);
        fadeout.keyOff(track);
        autoVib.keyOff(track);
    }
    
	public void newNote(double note, int track)
	{
		volEnv.newNote(note, track);
		panEnv.newNote(note, track);
		fadeout.newNote(note, track);
        autoVib.newNote(note, track);
	}

	public void reset(int track)
	{
		volEnv.reset(track);
		panEnv.reset(track);
		fadeout.reset(track);
        autoVib.reset(track);
	}

	public void setPosition(int pos, int track)
	{
		volEnv.setPosition(pos, track);
		panEnv.setPosition(pos, track);
	}
}
