/*
 *  (c) 2006 by VWP / Oxy
 *  (c) 2003 by Torkjel Hongve
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------
 */


package com.vwp.sound.mod.sound.output;

import java.io.IOException;

/**
 * @author torkjel
 *
 */
public interface Output
{
	/**
	 * the assumed sampeling rate of the input to write(byte[], int, int)
	 */
	public static final int INPUT_RATE = 44100; 


	/**
	 * tests if an output plugin is open for writeing
	 * @return true if the Output is open, false else 
	 */
	public boolean isOpen();

	/**
	 * opens the output plugin for writing
	 * @return true if the plugin was successfully opened, false else
	 */
	public boolean open();

	/**
	 * closes the output plugin
	 * @return true if the plugin was successfully closed, false else
	 */
	public boolean close();

	/**
	 * try to write len bytes from offset ofs from the array data.
	 * @param data the array containing data to write
	 * @param ofs the offset in data to write from
	 * @param len the number of samples to write
	 * @return the number of samples actually written
	 */
	public int write(byte[] data, int ofs, int len) throws IOException;
}
