/*
 *  (c) 2006 by VWP / Oxy
 *  (c) by Torkjel Hongve
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------
 */

package com.vwp.sound.mod.sound.output;

/**
 * A class representing the format of a stream of raw sound data. The sample rate,
 * number of bits per sample and the number of channels can be described.
 * @author torkjel
 *
 */
public class SoundDataFormat
{
	private int bits;
	private int rate;
	private int channels;

	/**
	 * @param bits the number of bits per sample, per channel (often 8 or 16)
	 * @param rate the sample rate (often 44100, 22050 or 11025)
	 * @param channels the number of channels (often 1, for mono, or 2, for stereo)
	 */
	public SoundDataFormat(int bits, int rate, int channels)
	{
		this.bits = bits;
		this.rate = rate;
		this.channels = channels;
	}

	/**
	 * @return the number of bits per sample
	 */
	public int getBits()
	{
		return bits;
	}

	/**
	 * @return the sample rate
	 */
	public int getRate()
	{
		return rate;
	}

	/**
	 * @return the number of channels
	 */
	public int getChannels()
	{
		return channels;
	}

	public String toString() {
	    return this.getClass().getName() + "[" +
            "bits=" + getBits() + ":" +
            "rate=" + getRate() + ":" +
            "channels=" + getChannels() + "]";
    }
}
